#include "DevAdcSeqBaseConfig.h"
#include "lqmeas/EnumConfigKey.h"
#include "lqmeas/StdConfigKeys.h"
#include <QJsonObject>
#include <QJsonArray>

namespace LQMeas {
    static const QLatin1String cfgkey_channels_array    {StdConfigKeys::channels()};
    static const QLatin1String cfgkey_ch_enabled        {StdConfigKeys::enabled()};
    static const QLatin1String cfgkey_freq_adc          {"FreqAdc"};
    static const QLatin1String cfgkey_freq_ch           {"FreqCh"};
    static const QLatin1String cfgkey_use_max_ch_freq   {"UseMaxFreqCh"};

    DevAdcSeqBaseConfig::DevAdcSeqBaseConfig(const DevAdcInfo &info) :
        DevAdcConfig{info},
        m_ch_cnt{info.adcChannelsCnt()},
        m_ch_enabled{new bool[m_ch_cnt]} {

        memset(&m_adc_params, 0, sizeof(m_adc_params));
        m_adc_params.FreqAdc = adcDefaultFreq();
        m_adc_params.UseMaxChFreq = true;
        memset(m_ch_enabled, 0, sizeof(m_ch_enabled[0])*static_cast<unsigned>(m_ch_cnt));
        m_results.AdcFreq = m_results.AdcChFreq = m_adc_params.FreqAdc;
    }

    DevAdcSeqBaseConfig::DevAdcSeqBaseConfig(const DevAdcSeqBaseConfig &cfg) :
        DevAdcConfig{cfg},
        m_ch_cnt{cfg.m_ch_cnt},
        m_ch_enabled{new bool[m_ch_cnt]} {

        memcpy(m_ch_enabled, cfg.m_ch_enabled, sizeof(m_ch_enabled[0])*static_cast<unsigned>(m_ch_cnt));
        memcpy(&m_adc_params, &cfg.m_adc_params, sizeof(m_adc_params));
        memcpy(&m_results, &cfg.m_results, sizeof(m_results));
    }

    DevAdcSeqBaseConfig::~DevAdcSeqBaseConfig() {
        delete [] m_ch_enabled;
    }

    void DevAdcSeqBaseConfig::adcSetChEnabled(int ch, bool en) {
        if (m_ch_enabled[ch] != en) {
            m_ch_enabled[ch] = en;
            notifyChEnableStateChanged(ch);
            notifyConfigChanged();
        }
    }

    void DevAdcSeqBaseConfig::adcSetFreq(double freq) {
        if (m_adc_params.FreqAdc != freq) {
            m_adc_params.FreqAdc = freq;
            notifyAdcFreqChanged();
            notifyConfigChanged();
        }
    }

    void DevAdcSeqBaseConfig::adcSetChFreq(double ch_freq) {
        if (m_adc_params.FreqCh != ch_freq) {
            m_adc_params.FreqCh = ch_freq;
            notifyChFreqChanged();
            notifyConfigChanged();
        }
    }

    void DevAdcSeqBaseConfig::adcSetUseMaxChFreq(bool en) {
        if (m_adc_params.UseMaxChFreq != en) {
            m_adc_params.UseMaxChFreq = en;
            notifyChFreqChanged();
            notifyConfigChanged();
        }
    }


    void DevAdcSeqBaseConfig::protUpdate() {
        DevAdcConfig::protUpdate();
        if ((m_adc_params.FreqAdc <= 0) || (m_adc_params.FreqAdc > adcInfo()->adcFreqMax())) {
            m_results.AdcFreq = adcInfo()->adcFreqMax();
        } else {
            m_results.AdcFreq = m_adc_params.FreqAdc;
        }

        if (m_adc_params.UseMaxChFreq || (m_adc_params.FreqCh <= 0) ||
                (m_adc_params.FreqCh > m_results.AdcFreq/adcEnabledChCnt())) {
            m_results.AdcChFreq = adcInfo()->adcFreqMax();
        } else {
            m_results.AdcChFreq = m_adc_params.FreqCh;
        }

        double prevAdcFreq = m_results.AdcFreq;
        double prevAdcChFreq = m_results.AdcChFreq;
        adcAdjustFreq(&m_results.AdcFreq, &m_results.AdcChFreq);

        if ((prevAdcFreq != m_results.AdcFreq) || (prevAdcChFreq != m_results.AdcChFreq)) {
            if (prevAdcFreq != m_results.AdcFreq) {
                notifyAdcFreqChanged();
            }
            if (prevAdcChFreq != m_results.AdcChFreq) {
                notifyChFreqChanged();
            }
            notifyConfigChanged();
        }
    }

    void DevAdcSeqBaseConfig::protSave(QJsonObject &adcObj) const     {
        QJsonArray chArray;
        for (int ch = 0; ch < m_ch_cnt; ++ch) {
            QJsonObject chObj;
            chObj[cfgkey_ch_enabled] = m_ch_enabled[ch];
            adcChSave(ch, chObj);
            chArray.append(chObj);
        }
        adcObj[cfgkey_channels_array] = chArray;
        adcObj[cfgkey_freq_adc] = m_adc_params.FreqAdc;
        adcObj[cfgkey_freq_ch] = m_adc_params.FreqCh;
        adcObj[cfgkey_use_max_ch_freq] = m_adc_params.UseMaxChFreq;
    }

    void DevAdcSeqBaseConfig::protLoad(const QJsonObject &adcObj) {
        const QJsonArray chArray = adcObj[cfgkey_channels_array].toArray();
        int ch_cnt = chArray.size();
        for (int ch=0; (ch < ch_cnt) && (ch < m_ch_cnt); ++ch) {
            const QJsonObject chObj = chArray.at(ch).toObject();
            m_ch_enabled[ch] = chObj[cfgkey_ch_enabled].toBool();
            adcChLoad(ch, chObj);
        }

        for (int ch=ch_cnt; ch < m_ch_cnt; ++ch) {
            m_ch_enabled[ch] = false;
            adcChSetDefault(ch);
        }

        double max_freq = adcInfo()->adcFreqMax();
        double adc_freq = adcObj[cfgkey_freq_adc].toDouble(adcDefaultFreq());
        if ((adc_freq > 0) && (adc_freq <= max_freq)) {
            m_adc_params.FreqAdc = adc_freq;
        } else {
            m_adc_params.FreqAdc = max_freq;
        }

        double chFreq = adcObj[cfgkey_freq_ch].toDouble();
        if ((chFreq >= 0) && (chFreq <= m_adc_params.FreqAdc)) {
            m_adc_params.FreqCh = chFreq;
        } else {
            m_adc_params.FreqCh = 0;
        }

        m_adc_params.UseMaxChFreq = adcObj[cfgkey_use_max_ch_freq].toBool(true);
    }
}
